# Generated by Django 5.2.3 on 2025-07-03 15:11
# noqa
import uuid

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="MalwareDetection",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        help_text="primary key for the record as UUID",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, help_text="date and time at which a record was created"),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, help_text="date and time at which a record was last updated"),
                ),
                ("path", models.CharField(help_text="path to the file", max_length=255, unique=True)),
                (
                    "status",
                    models.CharField(
                        choices=[("pending", "Pending"), ("processing", "Processing")],
                        default="pending",
                        help_text="status of the detection",
                        max_length=255,
                    ),
                ),
                (
                    "parameters",
                    models.JSONField(blank=True, default=dict(), help_text="parameters for the detection", null=True),
                ),
            ],
        ),
    ]
