"""Module contains the enums for the malware detection system."""

from enum import StrEnum


class ReportStatus(StrEnum):
    """The status of a report."""

    SAFE = "safe"
    UNSAFE = "unsafe"
    UNKNOWN = "unknown"
